/*
 *  linux/include/asm-arm/proc-armv/page.h
 *
 *  Copyright (C) 1995, 1996 Russell King
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */
#ifndef __ASM_PROC_PAGE_H
#define __ASM_PROC_PAGE_H

/* PAGE_SHIFT determines the page size */
#define PAGE_SHIFT	12

#define EXEC_PAGESIZE   4096

#endif /* __ASM_PROC_PAGE_H */
